﻿using System.Data;
using System.Runtime.InteropServices;
using System.Runtime.Serialization;
using System.Text;

public class Win32UserToken : IDisposable
{
    IntPtr _tokenInformation;
    TOKEN_USER? _tokenUser;
    string _sid = "";
    public string Sid => _sid;

    const uint ERROR_INSUFFICIENT_BUFFER = 122;

    [DllImport("advapi32.dll", SetLastError = true)]
    static extern bool GetTokenInformation(IntPtr TokenHandle, TOKEN_INFORMATION_CLASS TokenInformationClass, IntPtr TokenInformation, uint TokenInformationLength, out uint ReturnLength);

    [DllImport("advapi32", CharSet = CharSet.Auto, SetLastError = true)]
    static extern bool ConvertSidToStringSid(IntPtr pSID, out IntPtr ptrSid);

    public static implicit operator TOKEN_USER(Win32UserToken token)
    {
        if (token._tokenUser == null)
        {
            throw new NullReferenceException(nameof(token._tokenUser));
        }

        return token._tokenUser.Value;
    }

    public static unsafe bool TryParse(string sid, out Win32UserToken? instance)
    {
        instance = null;

        string [] ids = sid.Split('-');
        if (ids[0] != "S")
        {
            return false;
        }

        byte revision;
        if (byte.TryParse(ids[1], out revision) == false)
        {
            return false;
        }

        int identifierAuthority;
        if (int.TryParse(ids[2], out identifierAuthority) == false)
        {
            return false;
        }

        byte subAuthorityCount = (byte)(ids.Length - 3);
        int[] subAuthority = new int[subAuthorityCount];

        for (int i = 0; i < ids.Length - 3; i ++)
        {
            int.TryParse(ids[3 + i], out subAuthority[i]);
        }

        instance = new Win32UserToken();
        TOKEN_USER* pInstance = (TOKEN_USER*)Marshal.AllocHGlobal(0x18 + sizeof(int) * subAuthorityCount);
        pInstance->User.Sid = IntPtr.Add(new nint(pInstance), 0x10);
        pInstance->User.Attributes = 0;
        pInstance->Sid.Revision = revision;
        pInstance->Sid.IdentifierAuthorityAsValue = identifierAuthority;
        pInstance->Sid.SubAuthorityCount = subAuthorityCount;

        IntPtr ptr = IntPtr.Add(new IntPtr(pInstance), 0x18);
        for (int i = 0; i < subAuthorityCount; i++)
        {
            int* pInt = (int*)ptr.ToPointer();
            *(pInt) = subAuthority[i];            
            ptr = IntPtr.Add(ptr, 4);
        }

        instance = new Win32UserToken();
        instance._tokenInformation = new IntPtr(pInstance);
        instance.InitializeTokenFromMemory(instance._tokenInformation);

        return true;
    }

    private Win32UserToken() { }

    public Win32UserToken(IntPtr hToken)
    {
        _tokenInformation = GetTokenInformation(hToken);
        InitializeTokenFromMemory(_tokenInformation);
    }

    void InitializeTokenFromMemory(IntPtr hToken)
    { 
        _tokenUser = (TOKEN_USER?)Marshal.PtrToStructure(_tokenInformation, typeof(TOKEN_USER));
        if (_tokenUser == null)
        {
            return;
        }

        IntPtr pstr = IntPtr.Zero;
        if (ConvertSidToStringSid(_tokenUser.Value.User.Sid, out pstr) == true)
        {
            _sid = Marshal.PtrToStringAuto(pstr) ?? "";
            Marshal.FreeHGlobal(pstr);
        }
    }

    IntPtr GetTokenInformation(IntPtr hToken)
    {
        uint dwBufferSize;

        if (!GetTokenInformation(hToken, TOKEN_INFORMATION_CLASS.TokenUser, IntPtr.Zero, 0, out dwBufferSize))
        {
            int win32Result = Marshal.GetLastWin32Error();
            if (win32Result != ERROR_INSUFFICIENT_BUFFER)
            {
                return IntPtr.Zero;
            }
        }

        IntPtr tokenInformation = Marshal.AllocHGlobal((int)dwBufferSize);
        if (!GetTokenInformation(hToken, TOKEN_INFORMATION_CLASS.TokenUser, tokenInformation, dwBufferSize, out dwBufferSize))
        {
            return IntPtr.Zero;
        }

        return tokenInformation;
    }

    public unsafe string SubAuthority
    {
        get
        {
            if (this._tokenUser == null)
            {
                return "";
            }

            TOKEN_USER tokenUser = this._tokenUser.Value;

            // 만약 SubAuthorityCount의 수만큼 fixed SubAuthority 배열이 할당되었다면 이렇게 사용하는 것이 가능합니다.
            /*
            string[] texts = new string[tokenUser.Sid.SubAuthorityCount];
            for (int i = 0; i < tokenUser.Sid.SubAuthorityCount; i++)
            {
                int auth = tokenUser.Sid.SubAuthority[i];
                texts[i] = auth.ToString();
            }
            */

            IntPtr ptr = IntPtr.Add(this._tokenInformation, 0x18);
            string[] texts = new string[tokenUser.Sid.SubAuthorityCount];
            for (int i = 0; i < tokenUser.Sid.SubAuthorityCount; i++)
            {
                int auth = *(int *)ptr.ToPointer();
                texts[i] = auth.ToString();
                ptr = IntPtr.Add(ptr, 4);
            }

            return string.Join('-', texts);
        }
    }

    public override string ToString()
    {
        if (this._tokenUser == null)
        {
            return "";
        }

        TOKEN_USER tokenUser = this._tokenUser.Value;
        return $"S-{tokenUser.Sid.Revision}-{tokenUser.Sid.IdentifierAuthorityAsValue}-{this.SubAuthority}";
    }

    public void Dispose()
    {
        if (_tokenInformation != IntPtr.Zero)
        {
            Marshal.FreeHGlobal(_tokenInformation);
            _tokenInformation = IntPtr.Zero;
        }
    }
}

[StructLayout(LayoutKind.Sequential)]
public struct SID_AND_ATTRIBUTES
{
    public IntPtr Sid;
    public int Attributes;
}

[StructLayout(LayoutKind.Sequential)]
public struct TOKEN_USER
{
    public SID_AND_ATTRIBUTES User;
    public SecurityIdentifier Sid;
}

[StructLayout(LayoutKind.Sequential)]
public unsafe struct SecurityIdentifier
{
    public byte Revision;
    public byte SubAuthorityCount;
    public fixed byte IdentifierAuthority[6];
    public fixed int SubAuthority[1];

    public unsafe int IdentifierAuthorityAsValue
    {
        get
        {
            int result = 0;

            for (int i = 0; i < 6; i++)
            {
                byte value = IdentifierAuthority[i];
                result |= (value << (8 * (6 - (i + 1))));
            }

            return result;
        }

        set
        {
            int idAuthority = value & 0x3F;

            for (int i = 0; i < 6; i++)
            {
                long mask = 0xFF << (8 * i);
                long maskedValue = (idAuthority & mask);
                long idValue = maskedValue >> (8 * i);

                IdentifierAuthority[5 - i] = (byte)idValue;
            }
        }
    }

}

internal enum TOKEN_INFORMATION_CLASS
{
    TokenUser = 1,
}
